@@ -73,9 +73,10 @@ _xosd_lock(xosd * osd)
 {
   char c = 0;
   FUNCTION_START(Dlocking);
-  write(osd->pipefd[1], &c, sizeof(c));
-  pthread_mutex_lock(&osd->mutex);
-  FUNCTION_END(Dlocking);
+  if (write(osd->pipefd[1], &c, sizeof(c)) != -1) {
+    pthread_mutex_lock(&osd->mutex);
+  }
+    FUNCTION_END(Dlocking);
 }
 static /*inline */ void
 _xosd_unlock(xosd * osd)
@@ -83,11 +84,12 @@ _xosd_unlock(xosd * osd)
   char c;
   int generation = osd->generation, update = osd->update;
   FUNCTION_START(Dlocking);
-  read(osd->pipefd[0], &c, sizeof(c));
-  pthread_cond_signal(&osd->cond_wait);
-  pthread_mutex_unlock(&osd->mutex);
-  if (update & UPD_show)
-    _wait_until_update(osd, generation & ~1); /* no wait when already shown. */
+  if (read(osd->pipefd[0], &c, sizeof(c)) != -1) {
+    pthread_cond_signal(&osd->cond_wait);
+    pthread_mutex_unlock(&osd->mutex);
+    if (update & UPD_show)
+      _wait_until_update(osd, generation & ~1); /* no wait when already shown. */
+  }
   FUNCTION_END(Dlocking);
 }
 
@@ -202,43 +204,74 @@ draw_text(xosd * osd, int line)
   assert(osd);
   FUNCTION_START(Dfunction);
 
-  if (l->string == NULL)
-    return;
+  if (l->string != NULL) {
+    
+    if (l->width < 0) {
+      XRectangle rect;
+      XmbTextExtents(osd->fontset, l->string, strlen(l->string), NULL, &rect);
+      l->width = rect.width;
+    }
 
-  if (l->width < 0) {
-    XRectangle rect;
-    XmbTextExtents(osd->fontset, l->string, strlen(l->string), NULL, &rect);
-    l->width = rect.width;
-  }
+    switch (osd->align) {
+    case XOSD_center:
+      x = (osd->screen_width - l->width) / 2;
+      break;
+    case XOSD_right:
+      x = osd->screen_width - l->width - x;
+    case XOSD_left:
+      break;
+    }
 
-  switch (osd->align) {
-  case XOSD_center:
-    x = (osd->screen_width - l->width) / 2;
-    break;
-  case XOSD_right:
-    x = osd->screen_width - l->width - x;
-  case XOSD_left:
-    break;
-  }
-
-  if (osd->shadow_offset) {
-    XSetForeground(osd->display, osd->gc, osd->shadow_pixel);
-    _draw_text(osd, l->string, x + osd->shadow_offset,
-               y + osd->shadow_offset);
-  }
-  if (osd->outline_offset) {
-    int i, j;
-    XSetForeground(osd->display, osd->gc, osd->outline_pixel);
-    /* FIXME: echo . | osd_cat -O 50 -p middle -A center */
-    for (i = 1; i <= osd->outline_offset; i++)
-      for (j = 0; j < 9; j++)
-        if (j != 4)
-          _draw_text(osd, l->string, x + (j / 3 - 1) * i,
-                     y + (j % 3 - 1) * i);
-  }
-  if (1) {
-    XSetForeground(osd->display, osd->gc, osd->pixel);
-    _draw_text(osd, l->string, x, y);
+    if (osd->shadow_offset) {
+      XSetForeground(osd->display, osd->gc, osd->shadow_pixel);
+      if (osd->shadow_direction) {
+        switch(osd->shadow_direction) {
+          case 0:
+            _draw_text(osd, l->string, x, y - osd->shadow_offset);
+            break;
+          case 1:
+            _draw_text(osd, l->string, x + osd->shadow_offset, y - osd->shadow_offset);
+            break;
+          case 2:
+            _draw_text(osd, l->string, x + osd->shadow_offset, y);
+            break;
+          case 3:
+            _draw_text(osd, l->string, x + osd->shadow_offset, y + osd->shadow_offset);
+            break;
+          case 4:
+            _draw_text(osd, l->string, x, y + osd->shadow_offset);
+            break;
+          case 5:
+            _draw_text(osd, l->string, x - osd->shadow_offset, y + osd->shadow_offset);
+            break;
+          case 6:
+            _draw_text(osd, l->string, x - osd->shadow_offset, y);
+            break;
+          case 7:
+            _draw_text(osd, l->string, x - osd->shadow_offset, y - osd->shadow_offset);
+            break;
+          default:
+            //_draw_text(osd, l->string, x + osd->shadow_offset, y + osd->shadow_offset);
+            break;
+        }
+      } else {
+        _draw_text(osd, l->string, x +osd->shadow_offset, y + osd->shadow_offset);
+      }
+    }
+    if (osd->outline_offset) {
+      int i, j;
+      XSetForeground(osd->display, osd->gc, osd->outline_pixel);
+      /* FIXME: echo . | osd_cat -O 50 -p middle -A center */
+      for (i = 1; i <= osd->outline_offset; i++)
+        for (j = 0; j < 9; j++)
+          if (j != 4)
+            _draw_text(osd, l->string, x + (j / 3 - 1) * i,
+                      y + (j % 3 - 1) * i);
+    }
+    if (1) {
+      XSetForeground(osd->display, osd->gc, osd->pixel);
+      _draw_text(osd, l->string, x, y);
+    }
   }
 }
 
@@ -622,28 +655,55 @@ xosd_init(const char *font, const char *colour, int timeout, xosd_pos pos,
   xosd *osd = xosd_create(number_lines);
 
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return NULL;
-
   if (xosd_set_font(osd, font) == -1) {
     xosd_destroy(osd);
     /*
      * we do not set xosd_error, as set_font has already set it to 
      * a sensible error message. 
      */
-    return NULL;
+    return NULL; 
+  }
+  if(osd != NULL) {
+    xosd_set_colour(osd, colour);
+    xosd_set_timeout(osd, timeout);
+    xosd_set_pos(osd, pos);
+    xosd_set_vertical_offset(osd, voffset);
+    xosd_set_shadow_offset(osd, shadow_offset);
   }
-  xosd_set_colour(osd, colour);
-  xosd_set_timeout(osd, timeout);
-  xosd_set_pos(osd, pos);
-  xosd_set_vertical_offset(osd, voffset);
-  xosd_set_shadow_offset(osd, shadow_offset);
 
   return osd;
 }
 
 /* }}} */
 
+/* xosd_ -- Create a new xosd "object" with the same attributes as the provided xosd "object" {{{ */
+xosd *
+xosd_clone(xosd * osd2)
+{
+  xosd *osd = xosd_create(osd2->number_lines);
+  osd->align = osd2->align;
+  osd->bar_length = osd2->bar_length;
+  osd->shadow_colour = osd2->shadow_colour;
+  osd->shadow_pixel = osd2->shadow_pixel;
+/* Copying original lines to the cloned xosd instance causes unintuitive behaviour
+  osd->lines = osd2->lines; */
+  osd->pos = osd2->pos;
+  osd->hoffset = osd2->hoffset;
+  osd->voffset = osd2->voffset;
+  osd->colour = osd2->colour;
+  osd->pixel = osd2->pixel;
+  osd->outline_colour = osd2->outline_colour;
+  osd->outline_pixel = osd2->outline_pixel;
+  osd->shadow_offset = osd2->shadow_offset;
+  osd->shadow_direction = osd2->shadow_direction;
+  osd->outline_offset = osd2->outline_offset;
+  osd->screen_height = osd2->screen_height;
+  osd->screen_width = osd2->screen_width;
+  osd->screen_xpos = osd2->screen_xpos;
+  osd->nscreens = osd2->nscreens;
+  return osd;
+}
+
 /* xosd_create -- Create a new xosd "object" {{{ */
 xosd *
 xosd_create(int number_lines)
@@ -653,18 +713,13 @@ xosd_create(int number_lines)
   char *display;
   XSetWindowAttributes setwinattr;
   XGCValues xgcv = { .graphics_exposures = False };
-#ifdef HAVE_XINERAMA
-  int screens;
-  int dummy_a, dummy_b;
-  XineramaScreenInfo *screeninfo = NULL;
-#endif
 
   FUNCTION_START(Dfunction);
   DEBUG(Dtrace, "getting display");
   display = getenv("DISPLAY");
   if (!display) {
     xosd_error = "No display";
-    return NULL;
+    osd = NULL;
   }
 
   DEBUG(Dtrace, "Mallocing osd");
@@ -672,13 +727,13 @@ xosd_create(int number_lines)
   memset(osd, 0, sizeof(xosd));
   if (osd == NULL) {
     xosd_error = "Out of memory";
-    goto error0;
+    free(osd);
   }
 
   DEBUG(Dtrace, "Creating pipe");
   if (pipe(osd->pipefd) == -1) {
     xosd_error = "Error creating pipe";
-    goto error0b;
+    osd = NULL;
   }
 
   DEBUG(Dtrace, "initializing mutex");
@@ -693,7 +748,12 @@ xosd_create(int number_lines)
   osd->lines = malloc(sizeof(union xosd_line) * osd->number_lines);
   if (osd->lines == NULL) {
     xosd_error = "Out of memory";
-    goto error1;
+    pthread_cond_destroy(&osd->cond_sync);
+    pthread_cond_destroy(&osd->cond_wait);
+    pthread_mutex_destroy(&osd->mutex_sync);
+    pthread_mutex_destroy(&osd->mutex);
+    close(osd->pipefd[0]);
+    close(osd->pipefd[1]);
   }
 
   for (i = 0; i < osd->number_lines; i++)
@@ -715,14 +775,14 @@ xosd_create(int number_lines)
   osd->display = XOpenDisplay(display);
   if (!osd->display) {
     xosd_error = "Cannot open display";
-    goto error2;
+    free(osd->lines);
   }
   osd->screen = XDefaultScreen(osd->display);
 
   DEBUG(Dtrace, "x shape extension query");
   if (!XShapeQueryExtension(osd->display, &event_basep, &error_basep)) {
     xosd_error = "X-Server does not support shape extension";
-    goto error3;
+    XCloseDisplay(osd->display);
   }
 
   osd->visual = DefaultVisual(osd->display, osd->screen);
@@ -735,28 +795,13 @@ xosd_create(int number_lines)
      * if we still don't have a fontset, then abort 
      */
     xosd_error = "Default font not found";
-    goto error3;
+    XCloseDisplay(osd->display);
   }
 
   DEBUG(Dtrace, "width and height initialization");
-#ifdef HAVE_XINERAMA
-  if (XineramaQueryExtension(osd->display, &dummy_a, &dummy_b) &&
-      (screeninfo = XineramaQueryScreens(osd->display, &screens)) &&
-      XineramaIsActive(osd->display)) {
-    osd->screen_width = screeninfo[0].width;
-    osd->screen_height = screeninfo[0].height;
-    osd->screen_xpos = screeninfo[0].x_org;
-  } else
-#endif
-  {
-    osd->screen_width = XDisplayWidth(osd->display, osd->screen);
-    osd->screen_height = XDisplayHeight(osd->display, osd->screen);
-    osd->screen_xpos = 0;
-  }
-#ifdef HAVE_XINERAMA
-  if (screeninfo)
-    XFree(screeninfo);
-#endif
+
+  xosd_monitor(osd, 1);
+
   osd->line_height = 10 /*Dummy value */ ;
   osd->height = osd->line_height * osd->number_lines;
 
@@ -808,27 +853,146 @@ xosd_create(int number_lines)
   pthread_create(&osd->event_thread, NULL, event_loop, osd);
 
   return osd;
-
-error3:
-  XCloseDisplay(osd->display);
-error2:
-  free(osd->lines);
-error1:
-  pthread_cond_destroy(&osd->cond_sync);
-  pthread_cond_destroy(&osd->cond_wait);
-  pthread_mutex_destroy(&osd->mutex_sync);
-  pthread_mutex_destroy(&osd->mutex);
-  close(osd->pipefd[0]);
-  close(osd->pipefd[1]);
-error0b:
-  free(osd);
-error0:
-  return NULL;
 }
 
 /* }}} */
 
-/* xosd_uninit -- Destroy a xosd "object" {{{
+/* xosd_monitor -- swap the input xosd window's monitor parameters to correspond with desired screen */
+int 
+xosd_monitor(xosd * osd, int monitor)
+{
+   int return_value = -1;  
+   if (osd != NULL) {     
+    #ifdef HAVE_XINERAMA
+      int nscreens;
+      XineramaScreenInfo *screeninfo = NULL;
+      int dummy_a, dummy_b;
+    #endif
+    monitor--;
+
+    FUNCTION_START(Dfunction);
+
+    _xosd_lock(osd);
+
+    #ifdef HAVE_XINERAMA
+      if (XineramaQueryExtension(osd->display, &dummy_a, &dummy_b) &&
+          (screeninfo = XineramaQueryScreens(osd->display, &nscreens)) &&
+        XineramaIsActive(osd->display) && (nscreens > monitor) && (monitor >= 0)) {
+      osd->screen_width = screeninfo[monitor].width;
+      osd->screen_height = screeninfo[monitor].height;
+      osd->screen_xpos = screeninfo[monitor].x_org;
+      osd->nscreens = nscreens;
+      return_value = 0;
+      } else
+    #endif
+    {
+      osd->screen_width = XDisplayWidth(osd->display, osd->screen);
+      osd->screen_height = XDisplayHeight(osd->display, osd->screen);
+      osd->screen_xpos = 0;
+      xosd_error = "Error getting screen info from Xinerama";
+      return_value = -1;
+
+    }
+    #ifdef HAVE_XINERAMA
+      if (screeninfo) {
+        XFree(screeninfo);
+      }
+    #endif
+    osd->update |= UPD_pos;
+    _xosd_unlock(osd);
+  }
+  return return_value;
+}
+
+/* }}} */
+
+/* display_info_driver -- main for pthreads in display_info */
+void* 
+display_info_driver() 
+{
+  xosd * screen = xosd_create(1);
+  int nscreens = screen->nscreens;
+  int return_value = -1;
+  char word[256];
+  xosd * osd = malloc(sizeof(xosd*)*nscreens*3);
+  xosd ** osdptr = malloc(sizeof(xosd)*nscreens*3);
+  for (int i = 0; i < nscreens*3+1; i++) {
+    osdptr[i] = osd+i;
+  }
+  osdptr[0] = xosd_create(2);
+  if (osdptr[0] != NULL) {
+
+    FUNCTION_START(Dfunction);
+
+    xosd_set_outline_offset(osdptr[0], 1);
+    xosd_set_timeout(osdptr[0], 1);
+    xosd_set_font(osdptr[0], (char *) osd_default_font);
+      
+    for (int i = 0; i < nscreens; i++) {
+      osdptr[3*i] = xosd_clone(osdptr[0]);
+      xosd_monitor(osdptr[3*i], i+1);
+      xosd_set_align(osdptr[3*i], XOSD_center);
+      xosd_set_pos(osdptr[3*i], XOSD_middle);
+      sprintf(word, "%d", i+1);
+      xosd_display(osdptr[3*i], 0, XOSD_string, word);
+  
+      osdptr[3*i+1] = xosd_clone(osdptr[0]);
+      xosd_monitor(osdptr[3*i+1], i+1);
+      xosd_set_align(osdptr[3*i+1], XOSD_left);
+      xosd_set_pos(osdptr[3*i+1], XOSD_top);
+      sprintf(word, "%d", osdptr[i]->screen_width);
+      xosd_display(osdptr[3*i+1], 0, XOSD_string, word);
+      
+      osdptr[3*i+2] = xosd_clone(osdptr[0]);
+      xosd_monitor(osdptr[3*i+2], i+1);
+      xosd_set_align(osdptr[3*i+2], XOSD_right);
+      xosd_set_pos(osdptr[3*i+2], XOSD_top);
+      sprintf(word, "%d", osdptr[3*i+2]->screen_height);
+      xosd_display(osdptr[3*i+2], 0, XOSD_string, word);
+    }
+    sleep(15);
+    for (int i = 0; i < nscreens*3; i++) {
+      xosd_destroy(osdptr[i]);    
+    }
+    xosd_destroy(screen);
+    free(osd);
+    free(osdptr);
+    return_value = 0;
+    pthread_exit(&return_value);
+  }
+  xosd_destroy(screen);
+  pthread_exit(&return_value);
+}
+
+/* display_info -- graphically displays screen parameters */
+int 
+display_info()
+{
+  int i = -1;
+  pthread_t id;
+  pthread_create(&id, NULL, display_info_driver, &i);
+  return i;
+}
+
+/* }}} */
+
+/* screen_info -- returns the number of back-end screens on the X11 connection */
+int 
+screen_count(xosd * osd)
+{
+  int return_value;
+  if (osd->nscreens) {
+    return_value = osd->nscreens;
+  }
+  else {
+    return_value = -1;
+  }
+  return return_value;
+}
+
+/* }}} */
+
+/* xosd_uninit -- Destroy an xosd "object" {{{
  * Deprecated: Use xosd_destroy. */
 int
 xosd_uninit(xosd * osd)
@@ -839,53 +1003,52 @@ xosd_uninit(xosd * osd)
 
 /* }}} */
 
-/* xosd_destroy -- Destroy a xosd "object" {{{ */
+/* xosd_destroy -- Destroy an xosd "object" {{{ */
 int
 xosd_destroy(xosd * osd)
 {
   int i;
-
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
 
-  DEBUG(Dtrace, "waiting for threads to exit");
-  _xosd_lock(osd);
-  osd->done = 1;
-  _xosd_unlock(osd);
+    DEBUG(Dtrace, "waiting for threads to exit");
+    _xosd_lock(osd);
+    osd->done = 1;
+    _xosd_unlock(osd);
 
-  DEBUG(Dtrace, "join threads");
-  pthread_join(osd->event_thread, NULL);
+    DEBUG(Dtrace, "join threads");
+    pthread_join(osd->event_thread, NULL);
 
-  DEBUG(Dtrace, "freeing X resources");
-  XFreeGC(osd->display, osd->gc);
-  XFreeGC(osd->display, osd->mask_gc);
-  XFreeGC(osd->display, osd->mask_gc_back);
-  XFreePixmap(osd->display, osd->line_bitmap);
-  XFreeFontSet(osd->display, osd->fontset);
-  XFreePixmap(osd->display, osd->mask_bitmap);
-  XDestroyWindow(osd->display, osd->window);
+    DEBUG(Dtrace, "freeing X resources");
+    XFreeGC(osd->display, osd->gc);
+    XFreeGC(osd->display, osd->mask_gc);
+    XFreeGC(osd->display, osd->mask_gc_back);
+    XFreePixmap(osd->display, osd->line_bitmap);
+    XFreeFontSet(osd->display, osd->fontset);
+    XFreePixmap(osd->display, osd->mask_bitmap);
+    XDestroyWindow(osd->display, osd->window);
 
-  XCloseDisplay(osd->display);
+    XCloseDisplay(osd->display);
 
-  DEBUG(Dtrace, "freeing lines");
-  for (i = 0; i < osd->number_lines; i++)
-    if (osd->lines[i].type == LINE_text && osd->lines[i].text.string)
-      free(osd->lines[i].text.string);
-  free(osd->lines);
+    DEBUG(Dtrace, "freeing lines");
+    for (i = 0; i < osd->number_lines; i++)
+      if (osd->lines[i].type == LINE_text && osd->lines[i].text.string)
+        free(osd->lines[i].text.string);
+    free(osd->lines);
 
-  DEBUG(Dtrace, "destroying condition and mutex");
-  pthread_cond_destroy(&osd->cond_sync);
-  pthread_cond_destroy(&osd->cond_wait);
-  pthread_mutex_destroy(&osd->mutex_sync);
-  pthread_mutex_destroy(&osd->mutex);
-  close(osd->pipefd[0]);
-  close(osd->pipefd[1]);
+    DEBUG(Dtrace, "destroying condition and mutex");
+    pthread_cond_destroy(&osd->cond_sync);
+    pthread_cond_destroy(&osd->cond_wait);
+    pthread_mutex_destroy(&osd->mutex_sync);
+    pthread_mutex_destroy(&osd->mutex);
+    close(osd->pipefd[0]);
+    close(osd->pipefd[1]);
 
-  DEBUG(Dtrace, "freeing osd structure");
-  free(osd);
+    DEBUG(Dtrace, "freeing osd structure");
+    free(osd);
 
-  FUNCTION_END(Dfunction);
+    FUNCTION_END(Dfunction);
+  }
   return 0;
 }
 
@@ -896,17 +1059,13 @@ int
 xosd_set_bar_length(xosd * osd, int length)
 {
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-
-  if (length == 0)
-    return -1;
-  if (length < -1)
-    return -1;
-
-  osd->bar_length = length;
-
-  return 0;
+  int return_val = -1;
+  if (osd != NULL && length != 0 && length > 0) {
+    osd->bar_length = length;
+    return_val = 0;
+  } 
+  return return_val;
+  
 }
 
 /* }}} */
@@ -915,82 +1074,77 @@ xosd_set_bar_length(xosd * osd, int length)
 int
 xosd_display(xosd * osd, int line, xosd_command command, ...)
 {
-  int ret = -1;
-union xosd_line newline = { type:LINE_blank };
+  int return_value = -1;
+  union xosd_line newline = { type:LINE_blank };
   va_list a;
 
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-
-  if (line < 0 || line >= osd->number_lines) {
-    xosd_error = "xosd_display: Invalid Line Number";
-    return -1;
-  }
-
-  va_start(a, command);
-  switch (command) {
-  case XOSD_string:
-  case XOSD_printf:
-    {
-      char buf[XOSD_MAX_PRINTF_BUF_SIZE];
-      struct xosd_text *l = &newline.text;
-      char *string = va_arg(a, char *);
-      if (command == XOSD_printf) {
-        if (vsnprintf(buf, sizeof(buf), string, a) >= sizeof(buf)) {
-          xosd_error = "xosd_display: Buffer too small";
-          goto error;
+  if (osd != NULL && (line >= 0 && line < osd->number_lines)) {
+    va_start(a, command);
+    switch (command) {
+    case XOSD_string:
+    case XOSD_printf:
+      {
+        char buf[XOSD_MAX_PRINTF_BUF_SIZE];
+        struct xosd_text *l = &newline.text;
+        char *string = va_arg(a, char *);
+        if (command == XOSD_printf) {
+          if (vsnprintf(buf, sizeof(buf), string, a) >= sizeof(buf)) {
+            xosd_error = "xosd_display: Buffer too small";
+            va_end(a);
+            return_value = -1;
+          }
+          string = buf;
         }
-        string = buf;
+        if (string && *string) {
+          return_value = strlen(string);
+          l->type = LINE_text;
+          l->string = malloc(return_value + 1);
+          memcpy(l->string, string, return_value + 1);
+        } else {
+          return_value = 0;
+          l->type = LINE_blank;
+        }
+        l->width = -1;
+        break;
       }
-      if (string && *string) {
-        ret = strlen(string);
-        l->type = LINE_text;
-        l->string = malloc(ret + 1);
-        memcpy(l->string, string, ret + 1);
-      } else {
-        ret = 0;
-        l->type = LINE_blank;
+
+    case XOSD_percentage:
+    case XOSD_slider:
+      {
+        struct xosd_bar *l = &newline.bar;
+        return_value = va_arg(a, int);
+        return_value = (return_value < 0) ? 0 : (return_value > 100) ? 100 : return_value;
+        l->type = (command == XOSD_percentage) ? LINE_percentage : LINE_slider;
+        l->value = return_value;
+        break;
       }
-      l->width = -1;
+
+    default:
+      {
+        xosd_error = "xosd_display: Unknown command";
+        va_end(a);
+        return_value = -1;
+      }
+    }
+
+    _xosd_lock(osd);
+    /* Free old entry */
+    switch (osd->lines[line].type) {
+    case LINE_text:
+      free(osd->lines[line].text.string);
+    case LINE_blank:
+    case LINE_percentage:
+    case LINE_slider:
       break;
     }
+    osd->lines[line] = newline;
+    osd->update |= UPD_content | UPD_timer | UPD_show;
+    _xosd_unlock(osd);
 
-  case XOSD_percentage:
-  case XOSD_slider:
-    {
-      struct xosd_bar *l = &newline.bar;
-      ret = va_arg(a, int);
-      ret = (ret < 0) ? 0 : (ret > 100) ? 100 : ret;
-      l->type = (command == XOSD_percentage) ? LINE_percentage : LINE_slider;
-      l->value = ret;
-      break;
-    }
-
-  default:
-    {
-      xosd_error = "xosd_display: Unknown command";
-      goto error;
-    }
   }
-
-  _xosd_lock(osd);
-  /* Free old entry */
-  switch (osd->lines[line].type) {
-  case LINE_text:
-    free(osd->lines[line].text.string);
-  case LINE_blank:
-  case LINE_percentage:
-  case LINE_slider:
-    break;
-  }
-  osd->lines[line] = newline;
-  osd->update |= UPD_content | UPD_timer | UPD_show;
-  _xosd_unlock(osd);
-
-error:
-  va_end(a);
-  return ret;
+  
+  return return_value; 
 }
 
 /* }}} */
@@ -999,10 +1153,14 @@ error:
 int
 xosd_is_onscreen(xosd * osd)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-  return osd->generation & 1;
+  if (osd != NULL) {
+    return_val = osd->generation & 1;
+  }
+
+  return return_val;
 }
 
 /* }}} */
@@ -1012,15 +1170,18 @@ int
 xosd_wait_until_no_display(xosd * osd)
 {
   int generation;
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    return_val = 0;
+    if ((generation = osd->generation) & 1)
+      _wait_until_update(osd, generation);
 
-  if ((generation = osd->generation) & 1)
-    _wait_until_update(osd, generation);
+    FUNCTION_END(Dfunction);
+  }
 
-  FUNCTION_END(Dfunction);
-  return 0;
+  return return_val;
 }
 
 /* }}} */
@@ -1029,18 +1190,17 @@ xosd_wait_until_no_display(xosd * osd)
 int
 xosd_set_colour(xosd * osd, const char *colour)
 {
-  int retval = 0;
+  int return_val = -1;
 
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    return_val = parse_colour(osd, &osd->colour, &osd->pixel, colour);
+    osd->update |= UPD_lines;
+    _xosd_unlock(osd);
+  }
 
-  _xosd_lock(osd);
-  retval = parse_colour(osd, &osd->colour, &osd->pixel, colour);
-  osd->update |= UPD_lines;
-  _xosd_unlock(osd);
-
-  return retval;
+  return return_val;
 }
 
 /* }}} */
@@ -1049,18 +1209,17 @@ xosd_set_colour(xosd * osd, const char *colour)
 int
 xosd_set_shadow_colour(xosd * osd, const char *colour)
 {
-  int retval = 0;
+  int return_val = -1;
 
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    return_val = parse_colour(osd, &osd->shadow_colour, &osd->shadow_pixel, colour);
+    osd->update |= UPD_lines;
+    _xosd_unlock(osd);
+  }
 
-  _xosd_lock(osd);
-  retval = parse_colour(osd, &osd->shadow_colour, &osd->shadow_pixel, colour);
-  osd->update |= UPD_lines;
-  _xosd_unlock(osd);
-
-  return retval;
+  return return_val;
 }
 
 /* }}} */
@@ -1069,19 +1228,18 @@ xosd_set_shadow_colour(xosd * osd, const char *colour)
 int
 xosd_set_outline_colour(xosd * osd, const char *colour)
 {
-  int retval = 0;
+  int return_val = -1;
 
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    return_val =
+      parse_colour(osd, &osd->outline_colour, &osd->outline_pixel, colour);
+    osd->update |= UPD_lines;
+    _xosd_unlock(osd);
+  }
 
-  _xosd_lock(osd);
-  retval =
-    parse_colour(osd, &osd->outline_colour, &osd->outline_pixel, colour);
-  osd->update |= UPD_lines;
-  _xosd_unlock(osd);
-
-  return retval;
+  return return_val;
 }
 
 /* }}} */
@@ -1095,32 +1253,30 @@ xosd_set_font(xosd * osd, const char *font)
   char **missing;
   int nmissing;
   char *defstr;
-  int ret = 0;
+  int return_val = -1;
 
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-  if (font == NULL)
-    return -1;
-
-  /*
-   * Try to create the new font. If it doesn't succeed, keep old font. 
-   */
-  _xosd_lock(osd);
-  fontset2 = XCreateFontSet(osd->display, font, &missing, &nmissing, &defstr);
-  XFreeStringList(missing);
-  if (fontset2 == NULL) {
-    xosd_error = "Requested font not found";
-    ret = -1;
-  } else {
-    if (osd->fontset != NULL)
-      XFreeFontSet(osd->display, osd->fontset);
-    osd->fontset = fontset2;
-    osd->update |= UPD_font;
+  if (osd != NULL && font != NULL) {
+    /*
+    * Try to create the new font. If it doesn't succeed, keep old font. 
+    */
+    _xosd_lock(osd);
+    fontset2 = XCreateFontSet(osd->display, font, &missing, &nmissing, &defstr);
+    XFreeStringList(missing);
+    if (fontset2 == NULL) {
+      xosd_error = "Requested font not found";
+      return_val = -1;
+    } else {
+      if (osd->fontset != NULL)
+        XFreeFontSet(osd->display, osd->fontset);
+      osd->fontset = fontset2;
+      osd->update |= UPD_font;
+      return_val = 0;
+    }
+    _xosd_unlock(osd);
   }
-  _xosd_unlock(osd);
 
-  return ret;
+  return return_val;
 }
 
 /* }}} */
@@ -1129,38 +1285,56 @@ xosd_set_font(xosd * osd, const char *font)
 int
 xosd_set_shadow_offset(xosd * osd, int shadow_offset)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-  if (shadow_offset < 0)
-    return -1;
+  if (osd != NULL && shadow_offset > 0) {
+    _xosd_lock(osd);
+    osd->shadow_offset = shadow_offset;
+    osd->update |= UPD_font;
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
 
-  _xosd_lock(osd);
-  osd->shadow_offset = shadow_offset;
-  osd->update |= UPD_font;
-  _xosd_unlock(osd);
-
-  return 0;
+  return return_val;
 }
 
 /* }}} */
 
+/* xosd_set_shadow_direction -- Change the direction of the text shadow {{{*/
+int
+xosd_set_shadow_direction(xosd * osd, int shadow_direction) 
+{
+  int return_val = -1;
+
+  FUNCTION_START(Dfunction);
+  if (osd != NULL && (shadow_direction >= 0 && shadow_direction < 7)) {
+    _xosd_lock(osd);
+    osd->shadow_direction = shadow_direction;
+    osd->update |= UPD_font;
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
+
+  return return_val;
+}
+
 /* xosd_set_outline_offset -- Change the offset of the text outline {{{ */
 int
 xosd_set_outline_offset(xosd * osd, int outline_offset)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-  if (outline_offset < 0)
-    return -1;
+  if (osd != NULL && outline_offset >= 0) {
+    _xosd_lock(osd);
+    osd->outline_offset = outline_offset;
+    osd->update |= UPD_font;
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
 
-  _xosd_lock(osd);
-  osd->outline_offset = outline_offset;
-  osd->update |= UPD_font;
-  _xosd_unlock(osd);
-
-  return 0;
+  return return_val;
 }
 
 /* }}} */
@@ -1169,16 +1343,18 @@ xosd_set_outline_offset(xosd * osd, int outline_offset)
 int
 xosd_set_vertical_offset(xosd * osd, int voffset)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    osd->voffset = voffset;
+    osd->update |= UPD_pos;
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
 
-  _xosd_lock(osd);
-  osd->voffset = voffset;
-  osd->update |= UPD_pos;
-  _xosd_unlock(osd);
-
-  return 0;
+  return return_val;
 }
 
 /* }}} */
@@ -1187,16 +1363,18 @@ xosd_set_vertical_offset(xosd * osd, int voffset)
 int
 xosd_set_horizontal_offset(xosd * osd, int hoffset)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    osd->hoffset = hoffset;
+    osd->update |= UPD_pos;
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
 
-  _xosd_lock(osd);
-  osd->hoffset = hoffset;
-  osd->update |= UPD_pos;
-  _xosd_unlock(osd);
-
-  return 0;
+  return return_val;
 }
 
 /* }}} */
@@ -1205,16 +1383,18 @@ xosd_set_horizontal_offset(xosd * osd, int hoffset)
 int
 xosd_set_pos(xosd * osd, xosd_pos pos)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    osd->pos = pos;
+    osd->update |= UPD_pos;
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
 
-  _xosd_lock(osd);
-  osd->pos = pos;
-  osd->update |= UPD_pos;
-  _xosd_unlock(osd);
-
-  return 0;
+  return return_val;
 }
 
 /* }}} */
@@ -1223,16 +1403,18 @@ xosd_set_pos(xosd * osd, xosd_pos pos)
 int
 xosd_set_align(xosd * osd, xosd_align align)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    osd->align = align;
+    osd->update |= UPD_content;   /* XOSD_right depends on text width */
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
 
-  _xosd_lock(osd);
-  osd->align = align;
-  osd->update |= UPD_content;   /* XOSD_right depends on text width */
-  _xosd_unlock(osd);
-
-  return 0;
+  return return_val;
 }
 
 /* }}} */
@@ -1241,18 +1423,21 @@ xosd_set_align(xosd * osd, xosd_align align)
 int
 xosd_get_colour(xosd * osd, int *red, int *green, int *blue)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  if (osd != NULL) {
+    if (red)
+      *red = osd->colour.red;
+    if (blue)
+      *blue = osd->colour.blue;
+    if (green)
+      *green = osd->colour.green;
 
-  if (red)
-    *red = osd->colour.red;
-  if (blue)
-    *blue = osd->colour.blue;
-  if (green)
-    *green = osd->colour.green;
+    return_val = 0;
+}
 
-  return 0;
+  return return_val;
 }
 
 /* }}} */
@@ -1261,14 +1446,18 @@ xosd_get_colour(xosd * osd, int *red, int *green, int *blue)
 int
 xosd_set_timeout(xosd * osd, int timeout)
 {
+  int return_val = -1;
+
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-  _xosd_lock(osd);
-  osd->timeout = timeout;
-  osd->update |= UPD_timer;
-  _xosd_unlock(osd);
-  return 0;
+  if (osd != NULL) {
+    _xosd_lock(osd);
+    osd->timeout = timeout;
+    osd->update |= UPD_timer;
+    _xosd_unlock(osd);
+    return_val = 0;
+  }
+
+  return return_val;
 }
 
 /* }}} */
@@ -1277,18 +1466,20 @@ xosd_set_timeout(xosd * osd, int timeout)
 int
 xosd_hide(xosd * osd)
 {
-  FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  int return_val = -1;
 
-  if (osd->generation & 1) {
-    _xosd_lock(osd);
-    osd->update &= ~UPD_show;
-    osd->update |= UPD_hide;
-    _xosd_unlock(osd);
-    return 0;
+  FUNCTION_START(Dfunction);
+  if (osd != NULL) {
+    if (osd->generation & 1) {
+      _xosd_lock(osd);
+      osd->update &= ~UPD_show;
+      osd->update |= UPD_hide;
+      _xosd_unlock(osd);
+      return_val = 0;
+    }
   }
-  return -1;
+
+  return return_val;
 }
 
 /* }}} */
@@ -1297,18 +1488,20 @@ xosd_hide(xosd * osd)
 int
 xosd_show(xosd * osd)
 {
-  FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
+  int return_val = -1;
 
-  if (~osd->generation & 1) {
-    _xosd_lock(osd);
-    osd->update &= ~UPD_hide;
-    osd->update |= UPD_show | UPD_timer;
-    _xosd_unlock(osd);
-    return 0;
+  FUNCTION_START(Dfunction);
+  if (osd != NULL) {
+    if (~osd->generation & 1) {
+      _xosd_lock(osd);
+      osd->update &= ~UPD_hide;
+      osd->update |= UPD_show | UPD_timer;
+      _xosd_unlock(osd);
+      return_val = 0;
+    }
   }
-  return -1;
+
+  return return_val;
 }
 
 /* }}} */
@@ -1319,31 +1512,31 @@ xosd_scroll(xosd * osd, int lines)
 {
   int i;
   union xosd_line *src, *dst;
+  int return_val = -1;
 
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-  if (lines <= 0 || lines > osd->number_lines)
-    return -1;
-
-  _xosd_lock(osd);
-  /* Clear old text */
-  for (i = 0, src = osd->lines; i < lines; i++, src++)
-    if (src->type == LINE_text && src->text.string) {
-      free(src->text.string);
-      src->text.string = NULL;
+  if (osd != NULL && (lines > 0 && lines <= osd->number_lines)) {
+    _xosd_lock(osd);
+    /* Clear old text */
+    for (i = 0, src = osd->lines; i < lines; i++, src++)
+      if (src->type == LINE_text && src->text.string) {
+        free(src->text.string);
+        src->text.string = NULL;
+      }
+    /* Move following lines forward */
+    for (dst = osd->lines; i < osd->number_lines; i++)
+      *dst++ = *src++;
+    /* Blank new lines */
+    for (; dst < src; dst++) {
+      dst->type = LINE_blank;
+      dst->text.string = NULL;
     }
-  /* Move following lines forward */
-  for (dst = osd->lines; i < osd->number_lines; i++)
-    *dst++ = *src++;
-  /* Blank new lines */
-  for (; dst < src; dst++) {
-    dst->type = LINE_blank;
-    dst->text.string = NULL;
+    osd->update |= UPD_content;
+    _xosd_unlock(osd);
+    return_val = 0;
   }
-  osd->update |= UPD_content;
-  _xosd_unlock(osd);
-  return 0;
+
+  return return_val;
 }
 
 /* }}} */
@@ -1352,11 +1545,13 @@ xosd_scroll(xosd * osd, int lines)
 int
 xosd_get_number_lines(xosd * osd)
 {
+  int return_val = -1;
   FUNCTION_START(Dfunction);
-  if (osd == NULL)
-    return -1;
-
-  return osd->number_lines;
+  if (osd != NULL) {
+    return_val = osd->number_lines;
+  }
+    
+  return return_val;
 }
 
 /* }}} */
